/**
 * A typed multidimensional array returned by `makeMatrix`.
 */
type Matrix<D extends number, T = unknown> = Brand<_Matrix<D, T, []>>;
/**
 * Recursively constructs a `Matrix` type and keeps track of the number of recursions.
 *
 * @remarks
 * When a `_Matrix` is created without knowing it's exact size or number of dimensions, such
 * as when creating one dynamically, it's impossible recursively construct a type for it.
 * Instead, we short circuit the type to resolve to `any[]`, to avoid the typescript compiler
 * throwing a `type instantiation is excessively deep...` error. This ensures an array of at
 * least one dimension is returned, but unfortunately can't provide more safety than that.
 */
type _Matrix<D extends number, T = unknown, RecursionCountArray extends number[] = []> = number extends D ? any[] : D extends RecursionCountArray["length"] ? T : _Matrix<D, T[], [...RecursionCountArray, D]>;
/**
 * An array of numbers with defined length, corresponding to points in a `Matrix`.
 *
 * @see {@link https://github.com/microsoft/TypeScript/issues/26223} for discussion and other
 * implementations, though this seems to be the only one that works properly for `makeMatrix`'s use case.
 *
 * @remarks
 * When a `Vector` is created without knowing the exact number of dimensions, such as when
 * creating one dynamically, we short circuit the type to resolve to `number[]` so that
 * `makeMatrix` can accept a dynamically created `dimensions` parameter.
 */
type Vector<D extends number> = Brand<number extends D ? number[] : [number, ...number[]] & {
    readonly length: D;
}>;
/**
 * Stops intellisense from expanding types and exposing their more complicated internals to users.
 *
 * @see {@link https://github.com/microsoft/TypeScript/issues/31940} for discussion
 */
type Brand<D> = D & {};

/**
 * Creates a matrix (multi-dimensional array) with your desired dimensions and values.
 *
 * If a number is passed as the `dimensions` parameter, a matrix with that number of dimensions
 * is returned. If an array of numbers is passed, the returned matrix have those exact dimensions.
 *
 * @param dimensions    The desired dimensions of the matrix. A number or array of numbers.
 * @param initialValues The initial value for each point in the matrix. Can also take a callback function—with an argument that resolves to the current point's vector—to fill the matrix dynamically. Defaults to `null`.
 *
 * @example
 * const twoDArray = makeMatrix(2); // A 2D array
 * const threeXThreeXTwoArray = makeMatrix([3, 3, 2]); // A 3x3x2 array
 * const twoDNumberArray = makeMatrix([3, 5], 0); // A 3x5 array containing 0
 * const twoDRandomNumberArray = makeMatrix([1, 4], () => Math.random()); // A 1x4 array containing random numbers
 * const twoDVectorArray = makeMatrix([3, 3], vector => vector.join()); // A 3x3 array containing each point's co-ordinate as a string
 */
declare const makeMatrix: <D extends number, T>(dimensions: D | Vector<D>, initialValues?: T | ((vector: Vector<D>) => T) | null) => Matrix<D, T>;

export { Matrix, makeMatrix as default, makeMatrix };
